/****************************************************/
/* Bibliotheque d'image 3d                          */
/* image_3d_float.c                                 */
/*                                                  */
/* Ecrit par : Daniel Lacroix (all rights reserved) */
/*                                                  */
/****************************************************/

#include <image_3d_float.h>

/**************************************************/
/* Ce moteur 3D souffre de beaucoup de limitation */
/* - ne supporte pas le clipping en Z.            */
/* - ne supporte pas les eclairages.              */
/* - a finir le texturage.                        */
/* - un gros travail a faire pour optimiser.      */
/**************************************************/

/* table des cosinus et sinus pour acceler */
float *cos_table = NULL;

#define cos_func(x) cos_table[x&0x3FF]
#define sin_func(x) cos_table[(x-256)&0x3FF]

/*********************************************************/
/* transforme le point psrc en pdst avec la matrice pmat */
/* il est possible d'utiliser le meme point pour pdst et */
/* psrc (il et alors ecrit sur lui meme).                */
void trans_pix(dot_3d *pdst, dot_3d *psrc, float *pmat)
{
  /* c'est le produit ligne colonne mais avec elimination des */
  /* multiplications par zero d'ou la forme 'bizarre'         */
  pdst->x = pmat[12];
  pdst->y = pmat[13];
  pdst->z = pmat[14];
      
  if(psrc->x != 0.0)
  {
    if(pmat[0] != 0.0) pdst->x += psrc->x * pmat[0];
    if(pmat[1] != 0.0) pdst->y += psrc->x * pmat[1];
    if(pmat[2] != 0.0) pdst->z += psrc->x * pmat[2];
  }
  if(psrc->y != 0.0)
  {
    if(pmat[4] != 0.0) pdst->x += psrc->y * pmat[4];
    if(pmat[5] != 0.0) pdst->y += psrc->y * pmat[5];
    if(pmat[6] != 0.0) pdst->z += psrc->y * pmat[6];
  }
  if(psrc->z != 0.0)
  {  
    if(pmat[8] != 0.0)  pdst->x += psrc->z * pmat[ 8];
    if(pmat[9] != 0.0)  pdst->y += psrc->z * pmat[ 9];
    if(pmat[10] != 0.0) pdst->z += psrc->z * pmat[10];
  }
}
/*********************************************************/

/**************************************************************************/
/* multiplie la matrice p1 par p2, la matrice resultante est pdst         */
/* ATTENTION, pdst ne doit pas pointer sur les meme matrices que p1 et p2 */
void mul_mat(float *pdst, float *p1, float *p2)
{
  /* premiere ligne */
  pdst[0] = (p1[0]*p2[0])+(p1[1]*p2[4])+(p1[2]*p2[ 8])+(p1[3]*p2[12]);
  pdst[1] = (p1[0]*p2[1])+(p1[1]*p2[5])+(p1[2]*p2[ 9])+(p1[3]*p2[13]);
  pdst[2] = (p1[0]*p2[2])+(p1[1]*p2[6])+(p1[2]*p2[10])+(p1[3]*p2[14]);
  pdst[3] = (p1[0]*p2[3])+(p1[1]*p2[7])+(p1[2]*p2[11])+(p1[3]*p2[15]);
  /* deuxieme ligne */
  pdst[4] = (p1[4]*p2[0])+(p1[5]*p2[4])+(p1[6]*p2[ 8])+(p1[7]*p2[12]);
  pdst[5] = (p1[4]*p2[1])+(p1[5]*p2[5])+(p1[6]*p2[ 9])+(p1[7]*p2[13]);
  pdst[6] = (p1[4]*p2[2])+(p1[5]*p2[6])+(p1[6]*p2[10])+(p1[7]*p2[14]);
  pdst[7] = (p1[4]*p2[3])+(p1[5]*p2[7])+(p1[6]*p2[11])+(p1[7]*p2[15]);
  /* troisieme ligne */
  pdst[8]  = (p1[8]*p2[0])+(p1[9]*p2[4])+(p1[10]*p2[ 8])+(p1[11]*p2[12]);
  pdst[9]  = (p1[8]*p2[1])+(p1[9]*p2[5])+(p1[10]*p2[ 9])+(p1[11]*p2[13]);
  pdst[10] = (p1[8]*p2[2])+(p1[9]*p2[6])+(p1[10]*p2[10])+(p1[11]*p2[14]);
  pdst[11] = (p1[8]*p2[3])+(p1[9]*p2[7])+(p1[10]*p2[11])+(p1[11]*p2[15]);
  /* quatrieme ligne */
  pdst[12] = (p1[12]*p2[0])+(p1[13]*p2[4])+(p1[14]*p2[ 8])+(p1[15]*p2[12]);
  pdst[13] = (p1[12]*p2[1])+(p1[13]*p2[5])+(p1[14]*p2[ 9])+(p1[15]*p2[13]);
  pdst[14] = (p1[12]*p2[2])+(p1[13]*p2[6])+(p1[14]*p2[10])+(p1[15]*p2[14]);
  pdst[15] = (p1[12]*p2[3])+(p1[13]*p2[7])+(p1[14]*p2[11])+(p1[15]*p2[15]);    
}
/**************************************************************************/

/******************************************************************/
/* Renvoi le "dot product". Si les 2 vecteurs sont normalise cela */
/* revient a renvoyer le cosinus de l'angle entre les 2 vecteurs. */
float dot_product(dot_3d *p1, dot_3d *p2)
{ return(p1->x*p2->x + p1->y*p2->y + p1->z*p2->z); }
/******************************************************************/

/**************************************************************/
/* calcul le vecteur normal pdst au plan forme de p1,p2 et p3 */
/* pdst est normalise                                         */
void normal(dot_3d *pdst, dot_3d *p1, dot_3d *p2, dot_3d *p3)
{ float lenght;

  /* c'est le calcul du 'Cross Product' */
  pdst->x = (p2->y - p1->y) * (p3->z - p1->z) - (p2->z - p1->z) * (p3->y - p1->y);
  pdst->y = (p2->z - p1->z) * (p3->x - p1->x) - (p2->x - p1->x) * (p3->z - p1->z);
  pdst->z = (p2->x - p1->x) * (p3->y - p1->y) - (p2->y - p1->y) * (p3->x - p1->x);

  /* taille du vecteur normal */
  lenght = sqrt((pdst->x * pdst->x) + (pdst->y * pdst->y) + (pdst->z * pdst->z));

  /* on normalise la normal (taille egal a 1.0) */
  pdst->x /= lenght;
  pdst->y /= lenght;
  pdst->z /= lenght;
}
/**************************************************************/

/***************************************************************/
/* genere une matrice de translation dans pmat pour translater */
/* du vecteur (0.0,0.0,0.0) vers vecteur                       */
void move_mat(dot_3d *vecteur, float *pmat)
{
  pmat[ 0] = 1.0; pmat[5] = 1.0; pmat[10] = 1.0; pmat[15] = 1.0;
  pmat[12] = vecteur->x; pmat[13] = vecteur->y; pmat[14] = vecteur->z;
  /* tout le reste a 0 */
  pmat[1] = 0.0; pmat[2] = 0.0; pmat[3] = 0.0; pmat[ 4] = 0.0; pmat[6] = 0.0;
  pmat[7] = 0.0; pmat[8] = 0.0; pmat[9] = 0.0; pmat[11] = 0.0;
}
/***************************************************************/

/***************************************************************************/
/* genere une matrice de zoom de factor. factor = 1.0 => aucun changement  */
void zoom_mat(float factor, float *pmat)
{
  pmat[ 0] = factor; pmat[5] = factor; pmat[10] = factor;
  pmat[15] = 1.0;
  /* tout le reste a 0 */
  pmat[ 1] = 0.0; pmat[ 2] = 0.0; pmat[ 3] = 0.0; pmat[ 4] = 0.0;
  pmat[ 6] = 0.0; pmat[ 7] = 0.0; pmat[ 8] = 0.0; pmat[ 9] = 0.0;
  pmat[11] = 0.0; pmat[12] = 0.0; pmat[13] = 0.0; pmat[14] = 0.0;
}
/***************************************************************************/

/*************************************************************/
/* genere une matrice de rotation autour de l'axe x          */
/* angle est compris entre 0 (0 degres) et 1023 (359 degres) */
void x_rot_mat(int32 angle, float *pmat)
{
  pmat[ 5] =  cos_func(angle);
  pmat[ 6] =  sin_func(angle);
  pmat[ 9] = -sin_func(angle);
  pmat[10] =  cos_func(angle);
  pmat[ 0] = 1.0; pmat[15] = 1.0;
  /* tout le reste avec des 0 */
  pmat[ 1] = 0.0; pmat[ 2] = 0.0; pmat[ 3] = 0.0; pmat[ 4] = 0.0;
  pmat[ 7] = 0.0; pmat[ 8] = 0.0; pmat[11] = 0.0; pmat[12] = 0.0;
  pmat[13] = 0.0; pmat[14] = 0.0;
}
/*************************************************************/

/*************************************************************/
/* genere une matrice de rotation autour de l'axe y          */
/* angle est compris entre 0 (0 degres) et 1023 (359 degres) */
void y_rot_mat(int32 angle, float *pmat)
{
  pmat[ 0] =  cos_func(angle);
  pmat[ 8] =  sin_func(angle);
  pmat[ 2] = -sin_func(angle);
  pmat[10] =  cos_func(angle);
  pmat[ 5] = 1.0; pmat[15] = 1.0;
  /* tout le reste avec des 0 */
  pmat[ 1] = 0.0; pmat[ 3] = 0.0; pmat[ 4] = 0.0; pmat[ 6] = 0.0;
  pmat[ 7] = 0.0; pmat[ 9] = 0.0; pmat[11] = 0.0; pmat[12] = 0.0;
  pmat[13] = 0.0; pmat[14] = 0.0;
}
/*************************************************************/

/*************************************************************/
/* genere une matrice de rotation autour de l'axe z          */
/* angle est compris entre 0 (0 degres) et 1023 (359 degres) */
void z_rot_mat(int32 angle, float *pmat)
{
  pmat[ 0] =  cos_func(angle);
  pmat[ 1] =  sin_func(angle);
  pmat[ 4] = -sin_func(angle);
  pmat[ 5] =  cos_func(angle);
  pmat[10] = 1.0; pmat[15] = 1.0;
  /* tout le reste avec des 0 */
  pmat[ 2] = 0.0; pmat[ 3] = 0.0; pmat[ 6] = 0.0; pmat[ 7] = 0.0;
  pmat[ 8] = 0.0; pmat[ 9] = 0.0; pmat[11] = 0.0; pmat[12] = 0.0;
  pmat[13] = 0.0; pmat[14] = 0.0;  
}
/*************************************************************/

/*********************************************************************/
/* ajoute un objet 3d a l'image 3d, l'objet n'est pas recopie, toute */
/* modification sur l'original et donc repercute. L'objet et donc    */
/* sa deallocation sont a la charge de l'appelant.                   */
void image_3d_prepend_obj(image_3d *pimage_3d, obj_3d *obj)
{ obj_list *list_new;
  uint32 compt;

  if((list_new = (obj_list *)malloc(sizeof(obj_list))) == NULL)
  { perror("malloc failed\n"); exit(1); }

  list_new->obj  = obj;
  list_new->next = pimage_3d->list.next;
  pimage_3d->list.next = list_new;
  
  /* calcul les normales si ce n'est pas fait */
  if(obj->normal == NULL)
  {
    /* alloue la place pour les vecteur normaux */
    if((obj->normal = (dot_3d *)malloc(sizeof(dot_3d)*obj->nb_face)) == NULL)
    { perror("malloc failed\n"); exit(1); }
    
    /* fait les calculs pour chaque face */
    for(compt = 0; compt < obj->nb_face; compt++)
    {
      normal(&(obj->normal[compt]),&(obj->dot[obj->face[compt].pt1]),
        &(obj->dot[obj->face[compt].pt2]),&(obj->dot[obj->face[compt].pt3]));
      obj->face[compt].normal = compt;
    }
  }

  /* si la memoire pour les points une fois transformes n'est pas alloue, 
     on l'alloue ici */  
  if(obj->dot_tmp == NULL)
  {
    if((obj->dot_tmp = (dot_3d *)malloc(obj->nb_dot*sizeof(dot_3d))) == NULL)
    { perror("malloc failed "); exit(1); }
  }
}
/*********************************************************************/

/************************************************************************/
/* choisi la matrice de transformation qui s'applique a tout les objets */
/* la matrice n'est pas recopie, toute modification de l'original et    */
/* donc repercute. Sa deallocation et aussi a la charge de l'appelant   */
void image_3d_set_global_mat(image_3d *pimage_3d, float *pmat)
{ pimage_3d->global_mat = pmat; }
/************************************************************************/

/******************************************/
/* rempli les tables de cosinus et sinus  */
void init_cos_sin(void)
{ int32  pos;

  if((cos_table=(float *)malloc(sizeof(float)*1024)) == NULL)
  { perror("malloc failed\n"); exit(1); }
  for(pos = 0; pos < 1024; pos++)
  { cos_table[pos] = cos((pos*2.0*M_PI)/1024.0); }
}
/******************************************/

/******************************/
/* cree une nouvelle image 3d */
image_3d *image_3d_new(int32 width, int32 height)
{ image_3d *vimage_3d;
  
  if(cos_table == NULL) init_cos_sin();
  
  /* allocation de la structure elle meme */ 
  if((vimage_3d=(image_3d *)malloc(sizeof(image_3d))) == NULL)
  { perror("malloc failed\n"); exit(1); }
  /* choisi la distance focal */
  vimage_3d->focal_x = DEFAULT_FOCAL_X; vimage_3d->focal_y = DEFAULT_FOCAL_Y;
  /* rempli les dimensions de l'image */
  vimage_3d->width = width; vimage_3d->height = height;
  /* allocation du zbuffer correspondant a l'image */  
  if((vimage_3d->zbuf=(float *)malloc(width*height*sizeof(float))) == NULL)
  { perror("malloc failed\n"); exit(1); }
  /* pas de buffer d'objet au debut */
  vimage_3d->obj_buf = NULL;
  /* pas d'objet au debut */
  vimage_3d->list.next = NULL;
  /* pas de transformation global au debut */
  vimage_3d->global_mat = NULL;
  /* lumiere ambiante par defaut (il faut bien que l'on voit qq chose) */
  vimage_3d->ambiante = 0.25;
  return(vimage_3d);
}
/******************************/

/******************************************************/
/* si wanted == TRUE alors on active le buffer objet, */
/* sinon on le deactive (l'obj_buf sert a savoir quel */
/* objet se trouve a un point de l'image grace a son  */
/* numero d'identifiant id).                          */
void image_3d_set_obj_buffer(image_3d *pimage_3d, int wanted)
{
  if(wanted == TRUE)
  {
    /* si on en veut un et qu'il n'y en a pas deja, on le cree */
    if(pimage_3d->obj_buf == NULL)
    {
      /* allocation du buffer d'objects correspondant a l'image */
      if((pimage_3d->obj_buf=(uint32 *)malloc(pimage_3d->width*pimage_3d->height*sizeof(uint32))) == NULL)
      { perror("malloc failed\n"); exit(1); }
    }
  } else {
    /* si il y en avait deja un et que l'on en veut plus, on le libere */
    if(pimage_3d->obj_buf != NULL)
    {
      free(pimage_3d->obj_buf);
      pimage_3d->obj_buf = NULL;
    }
  }
}
/******************************************************/

/***********************/
/* libere une image_3d */
void image_3d_free(image_3d *pimage_3d)
{ obj_list *list_pos,*list_pos_last;

  /* libere le zbuffer si il existe */
  if(pimage_3d->zbuf != NULL)    free(pimage_3d->zbuf);
  /* libere le buffer objet si il existe */
  if(pimage_3d->obj_buf != NULL) free(pimage_3d->obj_buf);
  /* libere la liste des objets */
  list_pos = pimage_3d->list.next;
  while(list_pos != NULL)
  {
    list_pos_last = list_pos;
    list_pos = list_pos->next;
    free(list_pos_last);
  }
  /* libere la structure meme de l'image 3d */
  free(pimage_3d);
}
/***********************/

/***************************/
/* reinitialise le obj_buf */
void clear_obj_buf(image_3d *pimage_3d)
{ uint32 compt;
  /* si il n'y a pas de buffer objet, on quitte */
  if(pimage_3d->obj_buf == NULL) return;
  /* place la valeur qui represent aucun objet dans tout l'obj buffer */
  for(compt=0;compt<pimage_3d->width*pimage_3d->height;compt++)
    pimage_3d->obj_buf[compt] = -1;
}
/***************************/

/***************************/
/* reinitialise le zbuffer */
void clear_zbuf(image_3d *pimage_3d)
{ uint32 compt;
  /* si il n'y a pas de zbuffer, on quitte */
  if(pimage_3d->zbuf == NULL) return;
  /* place la valeur la plus eloigne possible dans tout le zbuffer */
  for(compt=0;compt<pimage_3d->width*pimage_3d->height;compt++)
    pimage_3d->zbuf[compt] = FAR_Z;
}
/***************************/

/********************/
/* libere un obj_3d */
void obj_3d_free(obj_3d *pobj_3d)
{ 
  /* libere les definitions des polygones */
  if(pobj_3d->face != NULL) free(pobj_3d->face);
  /* libere les points qui definissent les polygones */
  if(pobj_3d->dot != NULL) free(pobj_3d->dot);
  /* libere les points qui definissent les polygones apres transformation */
  if(pobj_3d->dot_tmp != NULL) free(pobj_3d->dot_tmp);
  /* libere les vecteurs normaux attache aux polygones */
  if(pobj_3d->normal != NULL) free(pobj_3d->normal);
  /* libere l'objet lui meme */
  free(pobj_3d);
}
/********************/

/**************************************************************/
/* projection du point src sur l'ecran (coordonnees dans dst) */
void project(image_3d *pimage_3d, dot_2d *dst, dot_3d *src)
{
  if(src->z != 0.0)
  {
    dst->x = ((pimage_3d->focal_x * src->x) / src->z) + (pimage_3d->width  >> 1);
    dst->y = (pimage_3d->height >> 1) - ((pimage_3d->focal_y * src->y) / src->z);
  } else {
    /* pour eviter une division par 0, on considere que z = 1 */
    dst->x = (pimage_3d->focal_x * src->x) + (pimage_3d->width  >> 1);
    dst->y = (pimage_3d->height >> 1) - (pimage_3d->focal_y * src->y);
  }
}
/**************************************************************/

/* quelques variables d'environnement utilises comme des parametres */
pix    *buf  = NULL;      /* l'image           */
float  *zbuf = NULL;      /* le zbuffer        */
uint32 *obj_buf = NULL;   /* le buffer d'objet */
uint32 current_obj = 0;   /* le numero d'identifiant de l'objet en cours de dessin */
int32  current_width;     /* la largeur de l'image */
int32  current_height;    /* la hauteur de l'image */
int32  current_width_mul_height; /* largeur*hauteur */
float  *current_matrice;  /* la matrice de transformation a appliquer */
float  current_focal_x,current_focal_y; /* distance focal */
image_3d *current_image_3d;
image    *current_image;
/* pour les equations de plan */
float A,B,C,D;
/* pour le texturage */
dot_3d O,V,H;

static void draw_obj(obj_3d *pobj_3d);

/*********************************************************/
/* recalcule l'image 3d. Le rendu est place dans pimage. */
/* pimage DOIT avoir la meme taille que l'image 3d.      */
void image_3d_render(image *pimage, image_3d *pimage_3d)
{ obj_list *list_pos;
  float vmat[16];

  /* reinitialisation du zbuffer */
  clear_zbuf(pimage_3d);
  /* reinitialisation du buffer objets */
  clear_obj_buf(pimage_3d);

  /* mise en place de variable d'environnement */
  buf  = pimage->buf;
  zbuf = pimage_3d->zbuf;
  obj_buf = pimage_3d->obj_buf;
  current_width  = pimage->width;
  current_height = pimage->height;
  current_width_mul_height = current_width * current_height;
  current_focal_x = pimage_3d->focal_x;
  current_focal_y = pimage_3d->focal_y;

  current_image_3d = pimage_3d;
  current_image = pimage;

  list_pos = pimage_3d->list.next;
  while(list_pos != NULL)
  {
    /* calcul de la matrice de transformation pour cet objet */
    if(pimage_3d->global_mat != NULL)
    {
      if(list_pos->obj->obj_mat != NULL)
      {
        mul_mat(vmat, list_pos->obj->obj_mat, pimage_3d->global_mat);
        current_matrice = vmat;
      } else { current_matrice = pimage_3d->global_mat; }
    } else {
      if(list_pos->obj->obj_mat != NULL)
        current_matrice = list_pos->obj->obj_mat;
      else current_matrice =  NULL;
    }
    /* identifiant de l'objet pour l'obj_buf */
    current_obj = list_pos->obj->id;
    /* on dessine l'objet */
    draw_obj(list_pos->obj);
    /* passe a l'objet suivant */
    list_pos = list_pos->next;
  }
}
/*********************************************************/

static void triangle_flat(dot_3d p1, dot_3d p2, dot_3d p3, dot_3d n, color pcol);
static void triangle_texture(dot_3d p1, dot_3d p2, dot_3d p3, dot_3d n);

/*********************************************************/
/* dessine un objet 3d. Se sert des variables globales.  */
/* Cette procedure ne doit pas etre appele a l'exterieur */
/* de ce paquage.                                        */
static void draw_obj(obj_3d *pobj_3d)
{ uint32 xinc;
  dot_3d n;

  /* calcul tout les points apres transformation */
  for(xinc = 0; xinc < pobj_3d->nb_dot; xinc++)
    trans_pix(&pobj_3d->dot_tmp[xinc], &pobj_3d->dot[xinc], current_matrice);
      
  /* dessine tout les polygones de cet objet */
  for(xinc = 0; xinc < pobj_3d->nb_face; xinc++)
  {
    /*trans_pix(&n,  &pobj_3d->normal[pobj_3d->face[xinc].normal], current_matrice);*/
    normal(&n,&pobj_3d->dot_tmp[pobj_3d->face[xinc].pt1],
      &pobj_3d->dot_tmp[pobj_3d->face[xinc].pt2],
      &pobj_3d->dot_tmp[pobj_3d->face[xinc].pt3]);
    /* choisi le bon rendu en fonction de l'objet */
    switch(pobj_3d->render_type)
    {
      case FLAT :
        triangle_flat(pobj_3d->dot_tmp[pobj_3d->face[xinc].pt1],
          pobj_3d->dot_tmp[pobj_3d->face[xinc].pt2],
          pobj_3d->dot_tmp[pobj_3d->face[xinc].pt3],n,pobj_3d->face[xinc].col1);
        break;
      case TEXTURED :
        /* ESSAI TEMPORAIRE */
	triangle_texture(pobj_3d->dot_tmp[pobj_3d->face[xinc].pt1],
          pobj_3d->dot_tmp[pobj_3d->face[xinc].pt2],
          pobj_3d->dot_tmp[pobj_3d->face[xinc].pt3],n);
	break;
      default :
        printf("Mode de rendu 3D inconnu\n");
	exit(2);
    }
  }
}
/*********************************************************/

static void hline_flat(int32 x1, int32 x2, int32 y, color pcol);

/***********************************************************/
/* Dessine un triangle 3D rempli d'une couleur uni (pcol). */
/* Se sert des variables globales.                         */
static void triangle_flat(dot_3d p1, dot_3d p2, dot_3d p3, dot_3d n, color pcol)
{
  dot_2d d1,d2,d3,dtmp;
  int xd1,yd1,xd2,yd2,i;
  int32 start,end;

  /* calcul l'equation du plan (Ax + By + Cz + D = 0) */
  D = -p1.x*(p2.y*p3.z-p3.y*p2.z) - p2.x*(p3.y*p1.z-p1.y*p3.z) -
      p3.x*(p1.y*p2.z-p2.y*p1.z);

  /* permet de savoir si la face nous tourne le dos.             */
  /* pour cela il faut que les points des polygones soit         */
  /* definie dans le sens de rotation des aiguilles d'une montre */
  if(D < 0.0) return;
  
  /* on termine le calcul de l'equation du plan */      
  A = p1.y*(p2.z-p3.z) + p2.y*(p3.z-p1.z) + p3.y*(p1.z-p2.z);
  B = p1.z*(p2.x-p3.x) + p2.z*(p3.x-p1.x) + p3.z*(p1.x-p2.x);
  C = p1.x*(p2.y-p3.y) + p2.x*(p3.y-p1.y) + p3.x*(p1.y-p2.y);
	
  /* on projete les points 3d sur l'ecran */
  project(current_image_3d, &d1, &p1);
  project(current_image_3d, &d2, &p2);
  project(current_image_3d, &d3, &p3);
  
  /* on tri les 3 points par ordre de y */
  if(d3.y < d2.y){ dtmp = d2; d2 = d3; d3 = dtmp; }
  if(d2.y < d1.y){ dtmp = d1; d1 = d2; d2 = dtmp; }
  if(d3.y < d2.y){ dtmp = d2; d2 = d3; d3 = dtmp; }

  xd1 = d2.x - d1.x;
  yd1 = d2.y - d1.y;
  xd2 = d3.x - d1.x;
  yd2 = d3.y - d1.y;
   
  if(yd1)
  {
    for(i = d1.y; i <= d2.y; i++)
    {
      start = d1.x + ((i - d1.y) * xd1) / yd1;
      end   = d1.x + ((i - d1.y) * xd2) / yd2;

      if(start < end)
        hline_flat(start,end,i,pcol);
      else
        hline_flat(end,start,i,pcol);
    }
  }
      
  xd1 = d3.x - d2.x;
  yd1 = d3.y - d2.y;
  
  if(yd1)
  {
    for(i = d2.y; i <= d3.y; i++)
    {
      start = d1.x + ((i - d1.y) * xd2) / yd2;
      end   = d2.x + ((i - d2.y) * xd1) / yd1;

      if(start < end)
        hline_flat(start,end,i,pcol);
      else
        hline_flat(end,start,i,pcol);
    }
  }
}
/***********************************************************/

/*********************************************************/
/* Dessine une ligne horizontale entre (x1,y) et (x2,y). */
/* x1 DOIT etre inferieur ou egal a x2.                  */
/* Se sert des variables globales.                       */
static void hline_flat(int32 x1, int32 x2, int32 y, color pcol)
{
  int32 x;
  int32 ligne;
  float z,dz;
  int32 rx1,rx2;
  
  if((y < 0) || (x2 < 0)) return;

  ligne = y * current_width;

  /* calcul les vrai x1 et x2 a cause du clipping (debordement = on recoupe)
     de l'image destination */
  rx1 = MAX(0,x1);
  rx2 = MIN(current_width-1,x2);
  
  for(x = rx1; x <= rx2; x++)
  {
    z = -D / ( ((A*(x-(current_width>>1))) / current_image_3d->focal_x) +
               ((B*((current_height>>1)-y)) / current_image_3d->focal_y) + C);

    if(z < zbuf[ligne + x])
    {
      zbuf[ligne + x] = z;
      buf[ligne + x].r = pcol.r;
      buf[ligne + x].g = pcol.g;
      buf[ligne + x].b = pcol.b;
    }
  }
}
/*********************************************************/

static void hline_texture(int32 x1, int32 x2, int32 y);

dot_3d ps1,ps2,ps3,ns;
dot_3d vec_light = {0,0,-1};
  
/***********************************************************/
/* Dessine un triangle 3D texture.                         */
/* Se sert des variables globales.                         */
static void triangle_texture(dot_3d p1, dot_3d p2, dot_3d p3, dot_3d n)
{
  dot_2d d1,d2,d3,dtmp;
  int32 xd1,yd1,xd2,yd2,i;
  int32 x1,x2,x3,y1,y2,y3;
  int32 start,end;
  float step1,step2;

  ps1 = p1; ps2 = p2; ps3 = p3; ns = n;

  /* calcul l'equation du plan (Ax + By + Cz + D = 0) */
  D = -p1.x*(p2.y*p3.z-p3.y*p2.z) - p2.x*(p3.y*p1.z-p1.y*p3.z) -
      p3.x*(p1.y*p2.z-p2.y*p1.z);

  /* permet de savoir si la face nous tourne le dos.             */
  /* pour cela il faut que les points des polygones soit         */
  /* definie dans le sens de rotation des aiguilles d'une montre */
  if(D < 0.0) return;
  
  /* on termine le calcul de l'equation du plan */      
  A = p1.y*(p2.z-p3.z) + p2.y*(p3.z-p1.z) + p3.y*(p1.z-p2.z);
  B = p1.z*(p2.x-p3.x) + p2.z*(p3.x-p1.x) + p3.z*(p1.x-p2.x);
  C = p1.x*(p2.y-p3.y) + p2.x*(p3.y-p1.y) + p3.x*(p1.y-p2.y);

/*
  printf("triangle (%f,%f,%f) (%f,%f,%f) (%f,%f,%f)\n",p1.x,p1.y,p1.z,
    p2.x,p2.y,p2.z,p3.x,p3.y,p3.z);
  printf("equation du plan : (%f)*x + (%f)*y + (%f)*z + (%f) = 0\n",A,B,C,D);
  
  printf("(%f,%f,%f).(%f,%f,%f) => light = %f\n",
    n.x,n.y,n.z,vec_light.x,vec_light.y,vec_light.z,dot_product(&n, &vec_light));
*/    
  /* on projete les points 3d sur l'ecran */
  project(current_image_3d, &d1, &p1);
  project(current_image_3d, &d2, &p2);
  project(current_image_3d, &d3, &p3);
  
  /* on tri les 3 points par ordre de y */
  if(d3.y < d2.y){ dtmp = d2; d2 = d3; d3 = dtmp; }
  if(d2.y < d1.y){ dtmp = d1; d1 = d2; d2 = dtmp; }
  if(d3.y < d2.y){ dtmp = d2; d2 = d3; d3 = dtmp; }

  x1 = d1.x; x2 = d2.x; x3 = d3.x;
  y1 = d1.y; y2 = d2.y; y3 = d3.y;

  xd1 = x2 - x1;
  yd1 = y2 - y1;
  xd2 = x3 - x1;
  yd2 = y3 - y1;
/*
  xd1 = d2.x - d1.x;
  yd1 = d2.y - d1.y;
  xd2 = d3.x - d1.x;
  yd2 = d3.y - d1.y;
*/

/*  if((int)(d3.y - d1.y) == 0) 
  {
    hline_texture(MIN3(d1.x,d2.x,d3.x),MAX3(d1.x,d2.x,d3.x),d1.y); return;
  }

  x2 = d1.x;
  step2 = (d3.x - d1.x)/(d3.y - d1.y);
  if((int)(d2.y - d1.y))
  {
    x1 = d1.x;
    step1 = (d2.x - d1.x)/(d2.y - d1.y);

    for(i = d1.y; i <= d2.y; i++)
    { start = x1; end = x2;
      x1 += step1; x2 += step2;
      if(start < end)
        hline_texture(start,end,i);
      else
        hline_texture(end,start,i);
    }
  } else {
    hline_texture(MIN(d1.x,d2.x),MAX(d1.x,d2.x),d1.y); return;
  }

  if((int)(d3.y - d2.y))
  {
    x1 = d2.x;  
    step1 = (d3.x - d2.x)/(d3.y - d2.y);

    for(i = d2.y; i <= d3.y; i++)
    { start = x1; end = x2;
      x1 += step1; x2 += step2;
      if(start < end)
        hline_texture(start,end,i);
      else
        hline_texture(end,start,i);
    }
  } else {
    hline_texture(MIN(d2.x,d3.x),MAX(d2.x,d3.x),d1.y); return;
  }
*/   
/*  if(yd1)
  {
    for(i = d1.y; i <= d2.y; i++)
    {
      start = d1.x + ((i - d1.y) * xd1) / yd1;
      end   = d1.x + ((i - d1.y) * xd2) / yd2;

      if(start < end)
        hline_texture(start,end,i);
      else
        hline_texture(end,start,i);
    }
  }
      
  xd1 = d3.x - d2.x;
  yd1 = d3.y - d2.y;
  
  if(yd1)
  {
    for(i = d2.y + 1; i <= d3.y; i++)
    {
      start = d1.x + ((i - d1.y) * xd2) / yd2;
      end   = d2.x + ((i - d2.y) * xd1) / yd1;

      if(start < end)
        hline_texture(start,end,i);
      else
        hline_texture(end,start,i);
    }
  }*/
  if(yd1)
  {
    for(i = y1; i < y2; i++)
    {
      start = x1 + ((i - y1) * xd1) / yd1;
      end   = x1 + ((i - y1) * xd2) / yd2;

      if(start < end)
        hline_texture(start,end,i);
      else
        hline_texture(end,start,i);
    }
  } /*else { hline_texture(MIN3(x1,x2,x3),MAX3(x1,x2,x3),y1); }*/
      
  xd1 = x3 - x2;
  yd1 = y3 - y2;
  
  if(yd1)
  {
    for(i = y2; i <= y3; i++)
    {
      start = x1 + ((i - y1) * xd2) / yd2;
      end   = x2 + ((i - y2) * xd1) / yd1;

      if(start < end)
        hline_texture(start,end,i);
      else
        hline_texture(end,start,i);
    }
  } /*else { hline_texture(MIN3(x1,x2,x3),MAX3(x1,x2,x3),y2); }*/
  
  if((y1 < current_height) && (y1 > 0) && (x1 < current_width) && (x1 > 0))
  {
    buf[y1*current_width+x1].r = 0xFF;
    buf[y1*current_width+x1].g = 0xFF;
    buf[y1*current_width+x1].b = 0;
  }

  if((y2 < current_height) && (y2 > 0) && (x2 < current_width) && (x2 > 0))
  {  
    buf[y2*current_width+x2].r = 0xFF;
    buf[y2*current_width+x2].g = 0xFF;
    buf[y2*current_width+x2].b = 0;
  }
  
  if((y3 < current_height) && (y3 > 0) && (x3 < current_width) && (x3 > 0))
  {
    buf[y3*current_width+x3].r = 0xFF;
    buf[y3*current_width+x3].g = 0xFF;
    buf[y3*current_width+x3].b = 0;
  }
}
/***********************************************************/

pix texture[16] = {{0xFF,0xD0,0xD0,0xFF},{0xFF,0xFF,0xFF,0xFF},{0xFF,0xFF,0xFF,0xFF},{0xFF,0xFF,0xFF,0xFF},
                   {0xFF,0xFF,0xFF,0xFF},{0xFF,0xD0,0xD0,0xFF},{0xFF,0xFF,0xFF,0xFF},{0xFF,0xFF,0xFF,0xFF},
		   {0xFF,0xFF,0xFF,0xFF},{0xFF,0xFF,0xFF,0xFF},{0x00,0xFF,0xFF,0xFF},{0xFF,0xFF,0xFF,0xFF},
		   {0xFF,0xFF,0xFF,0xFF},{0xFF,0xFF,0xFF,0xFF},{0xFF,0xFF,0xFF,0xFF},{0x00,0xFF,0xFF,0xFF}};

/*********************************************************/
/* Dessine une ligne horizontale entre (x1,y) et (x2,y). */
/* x1 DOIT etre inferieur ou egal a x2.                  */
/* Se sert des variables globales.                       */
static void hline_texture(int32 x1, int32 x2, int32 y)
{
  int32 x;
  int32 ligne;
  float z,dzinv,zinv,u,v;
  int32 rx1,rx2;
  dot_3d cdot;
  int ui,vi;
  float xr,yr;
  float light;

  /* quitte si la ligne n'est pas du tout visible */
  if((y < 0) || (y >= current_height) || (x2 < 0) || (x1 >= current_width)) return;

  ligne = y * current_width;

  /* calcul les vrai x1 et x2 a cause du clipping (debordement = on recoupe)
     de l'image destination */
  rx1 = MAX(0,x1);
  rx2 = MIN(current_width-1,x2);

  xr = x - (current_width>>1);
  yr = (current_height>>1) - y;
  
  for(x = rx1; x <= rx2; x++)
  {
    xr = x - (current_width>>1);

    z = -D / ( ((A*xr) / current_image_3d->focal_x) +
               ((B*yr) / current_image_3d->focal_y) + C);

    cdot.z = z;
    cdot.x = (xr * cdot.z) / current_image_3d->focal_x; 
    cdot.y = (yr * cdot.z) / current_image_3d->focal_y;
    
    /*printf("(%f,%f,%f)",cdot.x,cdot.y,cdot.z);*/
    
    /* u = (2xryjzj-xjyrzj-xjyjzr)/(2xiyjzj-xjyizj-xjyjzi) */
    u = (2*(cdot.x-ps1.x)*(ps3.y-ps1.y)*(ps3.z-ps1.z) -
         (ps3.x-ps1.x)*(cdot.y-ps1.y)*(ps3.z-ps1.z)-
	 (ps3.x-ps1.x)*(ps3.y-ps1.y)*(cdot.z-ps1.z))/
	(2*(ps2.x-ps1.x)*(ps3.y-ps1.y)*(ps3.z-ps1.z) -
	 (ps3.x-ps1.x)*(ps2.y-ps1.y)*(ps3.z-ps1.z) -
	 (ps3.x-ps1.x)*(ps3.y-ps1.y)*(ps2.z-ps1.z));

    /* v = (2xryizi-xiyrzi-xiyizr)/(2xjyizi-xiyjzi-xiyizj) */
    v = (2*(cdot.x-ps1.x)*(ps2.y-ps1.y)*(ps2.z-ps1.z) -
         (ps2.x-ps1.x)*(cdot.y-ps1.y)*(ps2.z-ps1.z)-
	 (ps2.x-ps1.x)*(ps2.y-ps1.y)*(cdot.z-ps1.z))/
	(2*(ps3.x-ps1.x)*(ps2.y-ps1.y)*(ps2.z-ps1.z) -
	 (ps2.x-ps1.x)*(ps3.y-ps1.y)*(ps2.z-ps1.z) -
	 (ps2.x-ps1.x)*(ps2.y-ps1.y)*(ps3.z-ps1.z));    

    u *= 4; v *= 4;

    light = dot_product(&ns, &vec_light);
    /* on y ajoute la lumiere ambiante qui baigne tout les objets */
    light += current_image_3d->ambiante;
    
    if(light < 0.0) light = 0.0;
    
    /*printf("(%f)",light);*/
    
    /*printf(">(%f,%f)",u,v);*/
    
    ui = ((int)u) & 3;
    vi = ((int)v) & 3;
    /*u = u % 4; v = v % 4;*/
    if(z < zbuf[ligne + x])
    {
      zbuf[ligne + x] = z;
      buf[ligne + x].r = (((float)texture[(vi*4) + ui].r)*light < 256.0)?(float)texture[(vi*4) + ui].r*light : 255;
      buf[ligne + x].g = (((float)texture[(vi*4) + ui].g)*light < 256.0)?(float)texture[(vi*4) + ui].g*light : 255;
      buf[ligne + x].b = (((float)texture[(vi*4) + ui].b)*light < 256.0)?(float)texture[(vi*4) + ui].b*light : 255;
    }
  }
  /*printf("\n");*/
}
/*********************************************************/

/* un tetraedre pour tester le moteur 3D */
dot_3d test_dot[4]  = { 0.0, 60.0,0.0, -40.0,-40.0, 0.0,
                       40.0,-40.0,0.0,   0.0,  0.0,60.0};
face   test_face[4] = {
   0,2,1,0,    {255,0,0},    {0,255,0},    {0,0,255},
   0,1,3,0,    {0,255,0},    {255,0,255},  {128,128,128},
   0,3,2,0,    {0,0,255},    {255,0,255},  {28,28,28},
   3,1,2,0,    {255,255,0},  {255,0,255},  {128,128,128}};
   
obj_3d test_obj = {
  /* mode de rendu */
  TEXTURED, /*FLAT,*/
  /* numero d'identifiant de l'objet (utilise dans obj_buf) */
  1,
  /* nombre de polygone de cet objet */
  4,
  /* definition de tout les polygones */
  test_face,
  /* nombre de point de cet objet */
  4,
  /* definition de tout les points (utilise dans les polygones) */
  test_dot,
  /* tampon pour les calculs des points de l'objet apres transformation */
  NULL,
  /* vecteur normaux (utilise dans les polygones) */
  NULL,
  /* transformation a appliquer sur l'objet (null si aucune) */
  NULL
};

