/****************************************************/
/* Bibliotheque d'image 3d                          */
/* image_3d_float.h                                 */
/*                                                  */
/* Ecrit par : Daniel Lacroix (all rights reserved) */
/*                                                  */
/****************************************************/

#ifndef __IMAGE_3D_FLOAT_H__
#define __IMAGE_3D_FLOAT_H__

#include <math.h>
#include <image.h>

/* une matrice est un tableau de 16 float, les matrices */
/* sont donc de 4 par 4. Les valeurs sont repartie a des*/
/* offset comme suit                                    */
/* | 0  1  2  3| */
/* | 4  5  6  7| */
/* | 8  9 10 11| */
/* |12 13 14 15| */

/* la 3d est definie d'apres un repere ou :            */
/* x est positif a droite                              */
/* y est positif en haut                               */
/* z est positif au loin (en s'enfoncant dans l'ecran) */

/* definition d'un point dans un repere 2d */
typedef struct {
  float x,y;
} dot_2d;

/* definition d'un point dans un repere 3d */
typedef struct {
  float x,y,z;
} dot_3d;

/* definition d'une couleur */
typedef struct {
  uint8 r,g,b;
} color;

/* definition d'un polygone */
typedef struct {
  uint32 pt1,pt2,pt3,normal;
  color  col1,col2,col3;
} face;

/* pour render_type */
enum{FLAT,TEXTURED};

/* definition d'un objet 3d */
typedef struct {
  uint8  render_type;
  /* numero d'identifiant de l'objet (utilise dans obj_buf) */
  uint32 id;
  /* nombre de polygone de cet objet */
  uint32 nb_face;
  /* definition de tout les polygones */
  face   *face;
  /* nombre de point de cet objet */
  uint32 nb_dot;
  /* definition de tout les points (utilise dans les polygones) */
  dot_3d *dot;
  /* tampon pour les calculs des points de l'objet apres transformation */
  dot_3d *dot_tmp;
  /* vecteur normaux (utilise dans les polygones) */
  dot_3d *normal;
  /* transformation a appliquer sur l'objet (null si aucune) */
  float  *obj_mat;
  
} obj_3d;

typedef struct _obj_list obj_list;

/* definition d'une liste d'objets 3d */
struct _obj_list {
  obj_3d   *obj;
  obj_list *next;
};

#define DEFAULT_FOCAL_X 120
#define DEFAULT_FOCAL_Y 120
/* valeur la plus eloigne possible */
#define FAR_Z           2E127

typedef struct {
  /* dimension de l'image 3d (utile pour la taille du zbuf et obj_buf) */
  int32 width,height;
  /* distance de la focal */
  float focal_x,focal_y;
  /* zbuffer pour decider si un point est cache ou pas */
  float  *zbuf;
  /* indique les objects visibles en chaque point. NULL si non utilise */
  uint32 *obj_buf;
  /* matrice de transformation a appliquer sur tout les objets (NULL si aucune) */
  float  *global_mat;
  /* liste des objets */
  obj_list list;
  /* pour la lumiere */
  float ambiante;
} image_3d;

/* transforme le point psrc en pdst avec la matrice pmat */
/* il est possible d'utiliser le meme point pour pdst et */
/* psrc (il et alors ecrit sur lui meme).                */
void trans_pix(dot_3d *pdst, dot_3d *psrc, float *pmat);

/* multipile la matrice p1 par p2, la matrice resultante est pdst         */
/* ATTENTION, pdst ne doit pas pointer sur les meme matrices que p1 et p2 */
void mul_mat(float *pdst, float *p1, float *p2);

/* calcul le vecteur normal pdst au plan forme de p1,p2 et p3 */
/* pdst est normalise a 2**NORMAL_SHL                         */
void normal(dot_3d *pdst, dot_3d *p1, dot_3d *p2, dot_3d *p3);

/* genere une matrice de translation dans pmat pour translater */
/* du vecteur (0,0,0) vers vecteur                             */
void move_mat(dot_3d *vecteur, float *pmat);

/* genere une matrice de zoom de factor. factor = 1.0 => aucun changement */
void zoom_mat(float factor, float *pmat);

/* genere une matrice de rotation autour de l'axe x          */
/* angle est compris entre 0 (0 degres) et 1023 (359 degres) */
void x_rot_mat(int32 angle, float *pmat);

/* genere une matrice de rotation autour de l'axe y          */
/* angle est compris entre 0 (0 degres) et 1023 (359 degres) */
void y_rot_mat(int32 angle, float *pmat);

/* genere une matrice de rotation autour de l'axe z          */
/* angle est compris entre 0 (0 degres) et 1023 (359 degres) */
void z_rot_mat(int32 angle, float *pmat);

/* cree une nouvelle image_3d */
image_3d *image_3d_new(int32 width, int32 height);

/* libere une image_3d */
void image_3d_free(image_3d *pimage_3d);

/* choisi la matrice de transformation qui s'applique a tout les objets */
/* la matrice n'est pas recopie, toute modification de l'original et    */
/* donc repercute. Sa deallocation et aussi a la charge de l'appelant   */
void image_3d_set_global_mat(image_3d *pimage_3d, float *pmat);

/* ajoute un objet 3d a l'image 3d, l'objet n'est pas recopie, toute */
/* modification sur l'original et donc repercute. L'objet et donc    */
/* sa deallocation son a la charge de l'appelant.                    */
void image_3d_prepend_obj(image_3d *pimage_3d, obj_3d *obj);

/* si wanted == TRUE alors on active le buffer objet, */
/* sinon on le deactive (l'obj_buf sert a savoir quel */
/* objet se trouve a un point de l'image grace a son  */
/* numero d'identifiant id).                          */
void image_3d_set_obj_buffer(image_3d *pimage_3d, int wanted);

/* libere un obj_3d */
void obj_3d_free(obj_3d *pobj_3d);

/* recalcule l'image 3d. Le rendu est place dans pimage. */
/* pimage DOIT avoir la meme taille que l'image 3d.      */
void image_3d_render(image *pimage, image_3d *pimage_3d);

/* quelques objets 3d pour faire des tests */
extern obj_3d test_obj;

#endif /* __IMAGE_3D_FLOAT_H__ */
