/**********************************************************************/
/* Bibliotheque de transformation d'images au format TrueColor 32bitd */
/* image_trans.c                                                      */
/*                                                                    */
/* Ecrit par : Daniel Lacroix (all rights reserved)                   */
/*                                                                    */
/**********************************************************************/

#include <image_trans.h>
#include <image_draw_func.h>
#include <math.h>
#include <stdio.h>

/***************************************************/
/* Redimentionne l'image psrc et place le resultat */
/* a la postion (0,0) dans pdest en prenant        */
/* toujours le point le plus proche.               */
void image_resize_nearest(image *pdest, image *psrc, int32 width, int32 height)
{ int32 vx,vy;
  pix   vcol;
  float rx,ry;
  float width_scale, height_scale;

  width_scale  = (float)psrc->width  / (float)width;
  height_scale = (float)psrc->height / (float)height;

  for(vy = 0;vy < height; vy++)  
  {
    for(vx = 0;vx < width; vx++)
    {
      rx = vx * width_scale;
      ry = vy * height_scale;
      vcol = get_pix(psrc, (int32)rx, (int32)ry);
      put_pix(pdest, vx, vy, vcol);
    }
  }
}
/***************************************************/


/***************************************************/
/* Redimentionne l'image psrc et place le resultat */
/* a la postion (0,0) dans pdest. On determine les */
/* points par interpolation bilinaire.             */
void image_resize_bilinear(image *pdest, image *psrc, int32 width, int32 height)
{ int32 vx,vy;
  pix   vcol,vcol1,vcol2,vcol3,vcol4;
  float rx,ry;
  float width_scale, height_scale;
  float x_dist, y_dist;

  width_scale  = (float)psrc->width  / (float)width;
  height_scale = (float)psrc->height / (float)height;

  for(vy = 0;vy < height; vy++)  
  {
    for(vx = 0;vx < width; vx++)
    {
      rx = vx * width_scale;
      ry = vy * height_scale;
      vcol1 = get_pix(psrc, (int32)rx, (int32)ry);
      vcol2 = get_pix(psrc, ((int32)rx)+1, (int32)ry);

      vcol3 = get_pix(psrc, (int32)rx, ((int32)ry)+1);
      vcol4 = get_pix(psrc, ((int32)rx)+1, ((int32)ry)+1);

      x_dist = rx - ((float)((int32)rx));
      y_dist = ry - ((float)((int32)ry));
      vcol.r = (vcol1.r*(1.0-x_dist) + vcol2.r*(x_dist))*(1.0-y_dist) + 
               (vcol3.r*(1.0-x_dist) + vcol4.r*(x_dist))*(y_dist);
      vcol.g = (vcol1.g*(1.0-x_dist) + vcol2.g*(x_dist))*(1.0-y_dist) + 
               (vcol3.g*(1.0-x_dist) + vcol4.g*(x_dist))*(y_dist);
      vcol.b = (vcol1.b*(1.0-x_dist) + vcol2.b*(x_dist))*(1.0-y_dist) + 
               (vcol3.b*(1.0-x_dist) + vcol4.b*(x_dist))*(y_dist);

      put_pix(pdest, vx, vy, vcol);
    }
  }
}
/***************************************************/

/***************************************************/
/* Redimentionne l'image psrc et place le resultat */
/* a la postion (0,0) dans pdest. On determine les */
/* points par interpolation avec une DCT 2D des 64 */
/* points les plus pres.                           */
void image_resize_dct(image *pdest, image *psrc, int32 width, int32 height)
{
  float dct_coeff[8][8];  /* 8 fonctions de bases definie par 8 valeurs */
  uint8 val[8][8];
  float dct_res[8][8];    /* coefficient 3d qui correspondent a nos pixels */
  float dc_coeff;
  
  /* initialise le tableau des fonctions de bases */
  double f1,f2,tmp;
  int x,y,coeff,i,i1,i2;

  boolean do_dct = TRUE;
  int32 vx,vy;
  float rx,ry,last_rx,last_ry;
  float width_scale, height_scale;
  float x_dist, y_dist;
  pix   vcol;
  
  uint32 nb_pix;
  float  percent_done,percent_tmp;

  void init_dct(void)
  {
    f1 = 2.0*atan(1.0)/8.0;
    for(y=0;y<8;y++){         /* de la fonction 0 a la 7              */
      f2=(double)y*f1;
      for(x=0;x<8;x++){       /* du point 0 au point 7 de la fonction */
        tmp=cos((double)(2*x+1)*f2);
        if(y==0) dct_coeff[y][x]=(float)(1.0/sqrt(2.0));
        else     dct_coeff[y][x]=(float)tmp;
      }
    }
    dc_coeff = 1.0/sqrt(2.0);
  }

  void calcul_dct(void)
  {    
    for(i2 = 0; i2 < 8; i2++)
    {
      for(i1 = 0; i1 < 8; i1++)
      {
        dct_res[i2][i1] = 0.0;

        for(y = 0; y < 8; y++)
          for(x = 0; x < 8; x++)
            dct_res[i2][i1] += dct_coeff[i1][x] * dct_coeff[i2][y] * (float)val[y][x];
        dct_res[i2][i1] /= 16.0;
      }
    }
  } /* fin de calcul_dct() */

  uint8 calcul_pix(float px, float py)
  { uint8 res;
  
    float temp;
    float dct_coeff_hor[8],dct_coeff_ver[8];

    for(y=0;y<8;y++)
    {
      f2=(double)y*f1;
      if(y==0) {
        dct_coeff_hor[y]=dc_coeff;
      } else {
        tmp=cos((double)(2.0*(4.0-px)+1.0)*f2);
        dct_coeff_hor[y]=(float)tmp;
      }
      if(y==0) {
        dct_coeff_ver[y]=dc_coeff;
      } else {
        tmp=cos((double)(2.0*(4.0-py)+1.0)*f2);
        dct_coeff_ver[y]=(float)tmp;
      }
    }
    
    temp = 0.0;
    for(y=0;y<8;y++)
      for(x=0;x<8;x++)
        temp += dct_coeff_hor[x] * dct_coeff_ver[y] * dct_res[y][x];

    if(temp > 255.0) res = 255;
    else if(temp < 0.0) res = 0;
    else res = (uint8)temp;
    return(res);
  }
  
  /* corp de la procedure */
  init_dct();
  
  width_scale  = (float)psrc->width  / (float)width;
  height_scale = (float)psrc->height / (float)height;

  nb_pix = 0;
  percent_done = 0.0;

  last_rx = last_ry = -1.0;
  
  for(vy = 0; vy < height; vy++)
  {
    for(vx = 0; vx < width; vx++)
    {
      rx = vx * width_scale;
      ry = vy * height_scale;

      x_dist = rx - ((float)((int32)rx));
      y_dist = ry - ((float)((int32)ry));
      
      if((((int32)last_rx) != ((int32)rx)) || (((int32)last_ry) != ((int32)ry)))
        do_dct = TRUE;
      else
        do_dct = FALSE;
      last_rx = rx;
      last_ry = ry;
      
      /* pour le rouge */
      for(y=0;y<8;y++)
        for(x=0;x<8;x++)
          val[y][x] = get_pix(psrc, ((int32)rx)-x+4, ((int32)ry)-y+4).r;
      calcul_dct();
      vcol.r = calcul_pix(x_dist, y_dist);
      /* pour le vert */
      for(y=0;y<8;y++)
        for(x=0;x<8;x++)
          val[y][x] = get_pix(psrc, ((int32)rx)-x+4, ((int32)ry)-y+4).g;
      calcul_dct();
      vcol.g = calcul_pix(x_dist, y_dist);
      /* pour le bleu */
      for(y=0;y<8;y++)
        for(x=0;x<8;x++)
          val[y][x] = get_pix(psrc, ((int32)rx)-x+4, ((int32)ry)-y+4).b;
      calcul_dct();
      vcol.b = calcul_pix(x_dist, y_dist);
      /* on place le pixel calcule */
      put_pix(pdest, vx, vy, vcol);
      
      nb_pix++;
      percent_tmp = (float)nb_pix / (float)(width*height);
      if(percent_tmp - percent_done >= 0.01)
      {
        percent_done = (float)(((uint32)(percent_tmp*100.0))/100.0);
        fprintf(stdout,"\rpercent done = %3.0f%c",percent_done*100.0,'%');
        fflush(stdout);
      }
    }
  }
  fprintf(stdout,"\n");
}
/***************************************************/

