/****************************************************/
/* Gestion de listes doublement chainees            */
/* list.c                                           */
/*                                                  */
/* Ecrit par : Daniel Lacroix (all rights reserved) */
/*                                                  */
/****************************************************/

#include <list.h>

/*********************************************************************/
/* Rajoute un element en tete                                        */
/* Si votre liste n'a pas d'element factice en tete, vous devez      */
/* recuperer le nouveau point d'entree de la liste qui est retourne. */
List *list_prepend(List *pList, pointer data)
{ List *vList;

  /* on alloue la memoire pour le nouvelle element */
  if((vList = (List *)malloc(sizeof(List))) == NULL)
  { perror("malloc failed "); exit(1); }

  if((pList != NULL) && (pList->data == NULL))
  { /* si il y a un element factice en tete */
    vList->data = data;
    vList->prev = pList;
    vList->next = pList->next;
    pList->next = vList;
    if(vList->next != NULL)
      vList->next->prev = vList; 
    return(pList);
  } else {
    /* si il n'y a pas d'element factice en tete */
    vList->data = data;
    vList->prev = NULL;
    vList->next = pList;
    if(pList != NULL)
      pList->prev = vList;
    return(vList);
  }
}
/*********************************************************************/

/*********************************************************************/
/* Rajoute un element en queue                                       */
/* Si votre liste n'a pas d'element factice en tete, vous devez      */
/* recuperer le nouveau point d'entree de la liste qui est retourne. */
List *list_append(List *pList, pointer data)
{ List *vList;
  List *vListMove,*vListMovePrev;

  /* on alloue la memoire pour le nouvelle element */
  if((vList = (List *)malloc(sizeof(List))) == NULL)
  { perror("malloc failed "); exit(1); }

  /* on se positionne a la fin de la liste */
  vListMovePrev = NULL;
  vListMove = pList;
  while(vListMove != NULL)
  { vListMovePrev = vListMove; vListMove = vListMove->next; }
  
  if(vListMovePrev == NULL)
  { /* si il n'y a pas d'element factice en tete et que la liste est vide */
    vList->data = data;
    vList->prev = NULL;
    vList->next = NULL;
    return(vList);
  } else {
    vList->data = data;
    vList->prev = vListMovePrev;
    vList->next = NULL;
    vListMovePrev->next = vList;
    return(pList);
  }
}
/*********************************************************************/

/******************************************************************/
/* Cree une nouvelle liste avec un element factice en tete.       */
/* Pour cree une liste sans element factice il suffit d'utiliser  */
/* un pointeur NULL. L'element factice permet d'avoir une adresse */
/* de debut de la liste qui ne change pas quand on insere des     */
/* elements dans la liste.                                        */
List *list_new(void)
{ List *vList;

  if((vList = (List *)malloc(sizeof(List))) == NULL)
  { perror("malloc failed "); exit(1); }

  /* l'element factice et caracterise par un data NULL */
  vList->data = NULL;
  vList->prev = NULL;
  vList->next = NULL;
  
  return(vList);
}
/******************************************************************/

/**************************************/
/* Libere la liste (pas les elements) */
void list_free(List *pList)
{ List *vListMove,*vListMoveNext;

  vListMove = pList;
  while(vListMove != NULL)
  { 
    vListMoveNext = vListMove->next;
    free(vListMove);
    vListMove = vListMoveNext;
  }
}
/**************************************/

/*************************************************************************/
/* Libere la liste et appele la procedure free_func pour chaque elements */
void list_free_full(List *pList, void (* free_func)(pointer data))
{ List *vListMove,*vListMoveNext;

  vListMove = pList;
  while(vListMove != NULL)
  {
    vListMoveNext = vListMove->next;
    free_func(vListMove->data);
    free(vListMove);
    vListMove = vListMoveNext;
  }
}
/*************************************************************************/

/***************************************************************************/
/* Libere la liste et appele la procedure free (du C) pour chaque elements */
void list_free_full_simple(List *pList)
{ List *vListMove,*vListMoveNext;

  vListMove = pList;
  while(vListMove != NULL)
  {
    vListMoveNext = vListMove->next;
    if(vListMove->data != NULL) free(vListMove->data);
    free(vListMove);
    vListMove = vListMoveNext;
  }
}
/***************************************************************************/

/*******************************************************/
/* Renvoi le nombre d'element contenue dans la liste   */
/* Si la liste a un fictif en tete alors il est compte */
uint32 list_nb_item(List *pList)
{ List *vListMove;
  uint32 compteur = 0;

  vListMove = pList;
  while(vListMove != NULL)
  { 
    if(vListMove->data != NULL) compteur++;
    vListMove = vListMove->next;
  }
  return(compteur);
}
/*******************************************************/

/************************************************************************************/
/* Libere un element de la liste (pas ce que l'element pointe si c'est un pointeur) */
List *list_free_item(List *pList, pointer data)
{ List *vListMove,*vListMoveNext;

  vListMove = pList;
  while((vListMove != NULL) && (vListMove->data != data))
  { vListMove = vListMove->next; }
  /* si on a trouve l'element, on le libere */
  if(vListMove != NULL)
  {
    if(vListMove->prev != NULL)
    {
      vListMove->prev->next = vListMove->next;
      if(vListMove->next != NULL)
        vListMove->next->prev = vListMove->prev;
      free(vListMove);
      return(pList);
    } else {
      vListMoveNext = vListMove->next;
      if(vListMoveNext != NULL)
        vListMoveNext->prev = NULL;
      free(vListMove);      
      return(vListMoveNext);
    }
  }
  return(pList);
}
/************************************************************************************/

/************************************************************************************/
/* Libere un element de la liste (pas ce que l'element pointe si c'est un pointeur) */
/* Si l'element a ete libere alors done == TRUE sinon FALSE                         */
List *list_free_item_with_check(List *pList, pointer data, boolean *done)
{ List *vListMove,*vListMoveNext;

  vListMove = pList;
  while((vListMove != NULL) && (vListMove->data != data))
  { vListMove = vListMove->next; }
  /* si on a trouve l'element, on le libere */
  if(vListMove != NULL)
  {
    *done = TRUE;
    if(vListMove->prev != NULL)
    {
      vListMove->prev->next = vListMove->next;
      if(vListMove->next != NULL)
        vListMove->next->prev = vListMove->prev;
      free(vListMove);
      return(pList);
    } else {
      vListMoveNext = vListMove->next;
      if(vListMoveNext != NULL)
        vListMoveNext->prev = NULL;
      free(vListMove);      
      return(vListMoveNext);
    }
  }
  *done = FALSE;
  return(pList);
}
/************************************************************************************/

/**************************************************************/
/* Renvoi l'element a la position item_nb dans la liste pList */
/* Si la liste a un fictif en tete alors il est compte        */
/* de meme que si l'element item_nb n'exite pas, la fonction  */
/* renvoi NULL. item_nb commence a 1.                         */
pointer list_item(List *pList, uint32 item_nb)
{ List *vListMove,*vListMoveNext;
  uint32 offset=0;

  vListMove = pList;
  while((vListMove != NULL) && (offset<item_nb-1))
  { vListMove = vListMove->next; offset++; }
  
  if(vListMove != NULL) return(vListMove->data);
  return(NULL);
}
/**************************************************************/

/**********************************************************************/
/* Renvoi l'element a la derniere position de la liste pList.         */
/* Renvoi NULL si la liste est vide ou le fictif si il n'y a que lui. */
List *list_last(List *pList)
{ List *vListMove;

  if(pList == NULL) return(NULL);
  vListMove = pList;
  while(vListMove->next != NULL)
  { vListMove = vListMove->next; }
  
  return(vListMove);
}
/**********************************************************************/

/*********************************************************************/
/* Rajoute un element apres l'element before. Si l'element before    */
/* n'est pas trouve alors le nouvel element sera rajoute a la fin    */
/* Si votre liste n'a pas d'element factice en tete, vous devez      */
/* recuperer le nouveau point d'entree de la liste qui est retourne. */
List *list_insert_after_item(List *pList, pointer before, pointer data)
{ List *vList;
  List *vListMove, *vListMovePrev;

  /* on alloue la memoire pour le nouvelle element */
  if((vList = (List *)malloc(sizeof(List))) == NULL)
  { perror("malloc failed "); exit(1); }

  vListMove = pList; vListMovePrev = NULL;
  while((vListMove != NULL) && (vListMove->data != before))
  { vListMovePrev = vListMove; vListMove = vListMove->next; }

  /* si on a trouve l'element precedent */
  if(vListMove != NULL)
  {
    vList->data = data;
    vList->prev = vListMove;
    vList->next = vListMove->next;
    vListMove->next = vList;
    /* met a jour le precedent du suivant */
    if(vList->next != NULL)
      vList->next->prev = vList;
    return(pList);
  /* si on n'a pas trouve l'element precedent */
  } else {
    if(vListMovePrev == NULL)
    { /* si il n'y a pas d'element factice en tete et que la liste est vide */
      vList->data = data;
      vList->prev = NULL;
      vList->next = NULL;
      return(vList);
    } else {
      vList->data = data;
      vList->prev = vListMovePrev;
      vList->next = NULL;
      vListMovePrev->next = vList;
      return(pList);
    }
  }
}
/*********************************************************************/
